
/**
 * @imports
 */
import _intersect from '@web-native-js/commons/arr/intersect.js';
import _with from '@web-native-js/commons/obj/with.js';
import cssRead from '../../css/readSync.js';
import TransformRule from '../../Css/TransformRule.js';
import getCanScroll from '../getCanScroll.js';
import getScrollParent from '../getScrollParent.js';
import calcIntersection from './calcIntersection.js.js.js.js';
import getOffsets from './getOffsets.js.js.js.js';
import Rect from '../geometry/Rect.js';

/**
 * Calculates the figures required to place an element
 * at another reference point in the UI.
 */
export default class Displacement {

    /**
     * Calculates the figures required to place an element
     * at another reference point in the UI and keeps the values for
     * further conversion to usable values.
     *
     * @param HTMLElement|Event|window 	el
     * @param HTMLElement|Event|window	reference
     * @param object 					options
     *
     * @return object
     */
    constructor(el, reference, options = {}) {
        this.el = el;
        this.reference = reference;
        this.options = options;
        this.intersection = Rect.of(this.options.using || this.el).intersectionWith(this.reference);
        this.alignment = {};
        this.coords = {};
        var length = {x:'width', y:'height'};
        var start = {x:'left', y:'top'};
        var end = {x:'right', y:'bottom'};
        ['x', 'y'].forEach(axis => {
            if (this.options[axis] === false) {
                return;
            }
            var rect1Length = this.intersection.rect1[length[axis]/*height*/];
            // Distinguish and predicate
            this.alignment[axis] = Displacement.parseDirective(this.options[axis] || '');
            switch(this.alignment[axis].keyword) {
                case 'before':
                    // Pull beyond start
                    this.coords[axis] = - (this.intersection[start[axis]/*top*/] + rect1Length);
                break;
                case 'after':
                    // Push beyond end
                    this.coords[axis] = this.intersection[end[axis]/*bottom*/] + rect1Length;
                break;
                case 'start':
                    // Pull to start
                    this.coords[axis] = - this.intersection[start[axis]/*top*/];
                break;
                case 'end':
                    // Push to end
                    this.coords[axis] = this.intersection[end[axis]/*bottom*/];
                break;
                default:
                    // Align to center
                    this.coords[axis] = this.intersection.delta[axis];
            }
            // Apply predicates
            if (this.alignment[axis].predicates) {
                this.coords[axis] += Displacement.evalDirectivePredicates(this.alignment[axis].predicates, rect1Length);
            }
        });
    }
    
    /**
    * Manipulates an element's translate.translate to place it with another element.
    *
    * @return object
    */
    get byTranslation() {
        if (!this.translateFormatted) {
            this.translateFormatted = {
                from: new TransformRule({translate: [],}),
                to: new TransformRule({translate: [],}),
                current: {
                    from: new TransformRule({translate: [],}),
                    to: new TransformRule({translate: [],}),
                    progress: {},
                },
            };
            // -----------
            var activeTransform = cssRead(this.el, 'transform');
            ['x', 'y'].forEach((axis, i) => {
                if (axis in this.coords) {
                    this.translateFormatted.from.translate[i] = 0;
                    this.translateFormatted.to.translate[i] = this.coords[axis] + activeTransform.translate[i];
                    // ----------------
                    this.translateFormatted.current.from.translate[i] = activeTransform.translate[i];
                    this.translateFormatted.current.to.translate[i] = this.coords[axis];
                    // ----------------
                    this.translateFormatted.current.progress[axis] = Math.abs(activeTransform.translate[i] / (this.coords[axis] + activeTransform.translate[i]));
                } else {
                    this.translateFormatted.from.translate[i] = activeTransform.translate[i];
                    this.translateFormatted.to.translate[i] = activeTransform.translate[i];
                    // ----------------
                    this.translateFormatted.current.from.translate[i] = activeTransform.translate[i];
                    this.translateFormatted.current.to.translate[i] = activeTransform.translate[i];
                    // ----------------
                    this.translateFormatted.current.progress[axis] = 0;
                }
            });
        }
       return this.translateFormatted;
    }
   
    /**
     * Calculates an element's offset where it to be at the given rect.
     *
     * @return object
     */
    get byOffset() {
        if (!this.offsetsFormatted) {
            var start = {x:'left', y:'top'};
            var end = {x:'right', y:'bottom'};
            var offsets = _intersect(['x', 'y'], Object.keys(this.coords)).reduce((obj, axis) => {
                switch(this.alignment[axis].keyword) {
                    case 'before':
                        return this.options.alternateAnchors ? _with(obj, start[axis], this.coords[axis]) : _with(obj, end[axis], - this.coords[axis]);
                    case 'after':
                        return this.options.alternateAnchors ? _with(obj, end[axis], - this.coords[axis]) : _with(obj, start[axis], this.coords[axis]);
                    case 'end':
                        return _with(obj, end[axis], - this.coords[axis]);
                    default: // center,start
                        return _with(obj, start[axis], this.coords[axis]);
                }
            }, {});
            // -----------
            this.offsetsFormatted = {from: {}, to: {}, current: {
                    from: {}, to: {}, progress: {},
                },
            };
            // -----------
            var currentOffsets = getOffsets(this.el, Object.keys(offsets));
            ['left', 'right', 'top', 'bottom'].forEach(name => {
                if (name in offsets) {
                    this.offsetsFormatted.from[name] = 0;
                    this.offsetsFormatted.to[name] = offsets[name] + currentOffsets[name];
                    // ----------------
                    this.offsetsFormatted.current.from[name] = currentOffsets[name];
                    this.offsetsFormatted.current.to[name] = offsets[name];
                    // ----------------
                    this.offsetsFormatted.current.progress[name] = Math.abs(currentOffsets[name] / (offsets[name] + currentOffsets[name]));
                } else {
                    this.offsetsFormatted.from[name] = 'auto';
                    this.offsetsFormatted.to[name] = 'auto';
                    // ----------------
                    this.offsetsFormatted.current.from[name] = 'auto';
                    this.offsetsFormatted.current.to[name] = 'auto';
                }
            });
        }
        return this.offsetsFormatted;
    }

    /**
     * Scrolls an element to the position of another element
     * within its scrollable parent.
     *
     * @return object
     */
    get byScroll() {
        if (!this.scrollingFormatted) {
            var viewport = this.options.scrollAnchor || getScrollParent(this.el);
            viewport = viewport === document.body ? window : viewport;
            if (!getCanScroll(viewport)) {
                return;
            }
            this.scrollingFormatted = {from: {}, to: {}, current: {
                    from: {}, to: {}, progress: {},
                },
                viewport: viewport,
            };
            // -----------
            var currentScroll = {
                scrollLeft: viewport[viewport === window ? 'pageXOffset' : 'scrollLeft'],
                scrollTop: viewport[viewport === window ? 'pageYOffset' : 'scrollTop'],
            };
            ['x', 'y'].forEach((axis, i) => {
                var axisProp = axis === 'x' ? 'scrollLeft' : 'scrollTop';
                if (axis in this.coords) {
                    this.scrollingFormatted.from[axisProp] = 0;
                    this.scrollingFormatted.to[axisProp] = this.coords[axis] + currentScroll[axisProp];
                    // ----------------
                    this.scrollingFormatted.current.from[axisProp] = currentScroll[axisProp];
                    this.scrollingFormatted.current.to[axisProp] = this.coords[axis];
                    // ----------------
                    this.scrollingFormatted.current.progress[axis] = Math.abs(currentScroll[axisProp] / (this.coords[axis] + currentScroll[axisProp]));
                } else {
                    this.scrollingFormatted.from[axisProp] = currentScroll[axisProp];
                    this.scrollingFormatted.to[axisProp] = currentScroll[axisProp];
                    // ----------------
                    this.scrollingFormatted.current.from[axisProp] = currentScroll[axisProp];
                    this.scrollingFormatted.current.to[axisProp] = currentScroll[axisProp];
                    // ----------------
                    this.scrollingFormatted.current.progress[axis] = 0;
                }
            });
        }
        return this.scrollingFormatted;
    }

    /**
     * Parses a directive to obtain a placement keyword and modifiers.
     *
     * @param string			 	expr
     *
     * @return object
     */
    static parseDirective(expr) {
        var regPlacement = new RegExp('(before|after|start|end|center)', 'g');
        var regModifiers = new RegExp('[\-\+][0-9]+(%)?', 'g');
        return {
            keyword: (expr.match(regPlacement) || [])[0],
            predicates: expr.match(regModifiers.replace(/ /g, '')),
        };
    }

    /**
     * Sums a list of Mathematical expressions.
     *
     * @param array				 	modifiers
     * @param number 				percentageContext
     *
     * @return number
     */
   static evalDirectivePredicates(modifiers, percentageContext) {
       return modifiers.reduce((total, modifier) => total + (modifier.endsWith('%') 
           ? parseFloat(modifier) / 100 * percentageContext
           : parseFloat(modifier)
       ), 0);
   }
};
